﻿using System;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Nutshell.Test.Support;
using System.Linq;
using System.Threading.Tasks;
using System.Threading;
using System.Collections.Generic;


namespace Nutshell.ThreadWorkerPool.Test
{
    [TestClass]
    public class ThreadWorkerPoolTest
    {
        [TestMethod]
        public void TestMinThreadWorkerCountSetting()
        {
            using (var manager = ThreadWorkerPoolManager.Instance)
            {
                ThreadWorkerPoolSettings settings = ThreadWorkerPoolSettings.Default;
                settings.MinThreadWorkerCount = 10;

                var pool = manager.CreateNewPool("TestPool", () => { }, settings);

                Assert.AreEqual(settings.MinThreadWorkerCount, pool.IdleThreadWorkerCount);
            }
        }

        [TestMethod]
        public void TestIdleTimeSetting()
        {
            using (var manager = ThreadWorkerPoolManager.Instance)
            {
                ThreadWorkerPoolSettings settings = ThreadWorkerPoolSettings.Default;
                settings.MinThreadWorkerCount = 1;
                settings.IdleTime = 1000;

                var pool = manager.CreateNewPool("TestPool", () => { }, settings);
                IThreadWorker worker1;
                IThreadWorker worker2;
                pool.TryTake(out worker1);
                pool.TryTake(out worker2);
                Assert.AreEqual(0, pool.IdleThreadWorkerCount);
                Assert.AreEqual(2, pool.TotalThreadWorkerCount);

                worker1.Work();
                worker2.Work();

                Thread.Sleep(settings.IdleTime + 10);

                Assert.AreEqual(1, pool.IdleThreadWorkerCount);
                Assert.AreEqual(1, pool.TotalThreadWorkerCount);
            }
        }

        [TestMethod]
        public void TestTryTakeTimeOut()
        {
            using (var manager = ThreadWorkerPoolManager.Instance)
            {
                ThreadWorkerPoolSettings settings = ThreadWorkerPoolSettings.Default;
                settings.MinThreadWorkerCount = 0;
                settings.MaxThreadWorkerCount = 1;

                var pool = manager.CreateNewPool("TestPool", () => { }, settings);

                int timeout = 1000;
                IThreadWorker worker;

                Assert.IsTrue(pool.TryTake(timeout, out worker));
                Assert.IsNotNull(worker);

                bool isTakeOne = false;
                Thread thread = new Thread(() =>
                {
                    isTakeOne = pool.TryTake(timeout, out worker);
                });
                thread.Start();

                Thread.Sleep(timeout + 10);

                Assert.IsFalse(isTakeOne);
                Assert.IsNull(worker);
            }
        }

        [TestMethod]
        public void TestWork()
        {
            using (var manager = ThreadWorkerPoolManager.Instance)
            {
                ThreadWorkerPoolSettings settings = ThreadWorkerPoolSettings.Default;
                settings.MinThreadWorkerCount = 3;
                settings.MaxThreadWorkerCount = 20;
                settings.IdleTime = 1000;
                System.Collections.Concurrent.ConcurrentBag<int> result = new System.Collections.Concurrent.ConcurrentBag<int>();
                var pool = manager.CreateNewPool<int>("TestPool2", (data) =>
                {
                    try
                    {
                        result.Add(data);
                    }
                    catch (Exception ex)
                    {
                        Console.WriteLine(ex.Message);
                    }
                }, settings);

                ParameterizedThreadStart testWork = (obj) =>
                {
                    IThreadWorker<int> worker;
                    int value = (int)obj;
                    for (int index = 0; index < 100; index++)
                    {
                        if (pool.TryTake(out worker))
                        {
                            worker.Work(value + index);
                        }
                    }
                };
                Thread[] threadList = new Thread[3];
                threadList[0] = new Thread(testWork);
                threadList[1] = new Thread(testWork);
                threadList[2] = new Thread(testWork);
                for (int index = 0; index < threadList.Length; index++)
                {
                    threadList[index].Start(index * 100 + 1);
                }

                for (int index = 0; index < threadList.Length; index++)
                {
                    threadList[index].Join();
                }

                while (pool.IdleThreadWorkerCount != pool.TotalThreadWorkerCount)
                {
                    Thread.Sleep(100);
                }

                Assert.AreEqual(threadList.Length * 100, result.Distinct().Count());

                Thread.Sleep(pool.Settings.IdleTime + 100);
                Assert.AreEqual(pool.Settings.MinThreadWorkerCount, pool.TotalThreadWorkerCount);
            }
        }
    }
}
