﻿using Nutshell.ThreadWorkerPool.Handlers;
using System;
using System.Collections.Generic;

namespace Nutshell.ThreadWorkerPool
{
    public sealed class ThreadWorkerPoolManager : IDisposable
    {
        private readonly static ThreadWorkerPoolManager _threadWorkerPoolManager = new ThreadWorkerPoolManager();
        private readonly Dictionary<string, ThreadWorkerPoolBase> _poolMappings;
        private readonly object _createLocker = new object();

        static ThreadWorkerPoolManager() { }
        ThreadWorkerPoolManager()
        {
            _poolMappings = new Dictionary<string, ThreadWorkerPoolBase>();
        }

        public static ThreadWorkerPoolManager Instance
        {
            get { return _threadWorkerPoolManager; }
        }

        public IThreadWorkerPool CreateNewPool(string poolName, WorkingHandler workingHandler, ThreadWorkerPoolSettings poolSettings)
        {
            if (string.IsNullOrWhiteSpace(poolName))
                throw new ArgumentNullException("poolName");

            if (workingHandler == null)
                throw new ArgumentNullException("workingHandler");

            ThreadWorkerPool pool;
            lock (_createLocker)
            {
                if (_poolMappings.ContainsKey(poolName))
                    throw new ArgumentException(string.Format("Name is {0} ThreadWorkerPool is exists!", poolName), "poolName");

                pool = new ThreadWorkerPool(poolName, workingHandler, poolSettings);
                pool.Init();
                _poolMappings.Add(poolName, pool);
            }

            return pool;
        }

        public IThreadWorkerPool<T> CreateNewPool<T>(string poolName, WorkingHandler<T> workingHandler, ThreadWorkerPoolSettings poolSettings)
        {
            if (string.IsNullOrWhiteSpace(poolName))
                throw new ArgumentNullException("poolName");

            if (workingHandler == null)
                throw new ArgumentNullException("workingHandler");

            ThreadWorkerPool<T> pool;
            lock (_createLocker)
            {
                if (_poolMappings.ContainsKey(poolName))
                    throw new ArgumentException(string.Format("Name is {0} ThreadWorkerPool is exists!", poolName), "poolName");

                pool = new ThreadWorkerPool<T>(poolName, workingHandler, poolSettings);
                pool.Init();
                _poolMappings.Add(poolName, pool);
            }

            return pool;
        }

        public IThreadWorkerPool Get(string poolName)
        {
            if (string.IsNullOrWhiteSpace(poolName))
                throw new ArgumentNullException("poolName");

            ThreadWorkerPoolBase pool = null;
            _poolMappings.TryGetValue(poolName, out pool);
            return pool as IThreadWorkerPool;
        }

        public IThreadWorkerPool<T> Get<T>(string poolName)
        {
            if (string.IsNullOrWhiteSpace(poolName))
                throw new ArgumentNullException("poolName");

            ThreadWorkerPoolBase pool = null;
            _poolMappings.TryGetValue(poolName, out pool);
            return pool as IThreadWorkerPool<T>;
        }

        public IEnumerable<ThreadWorkerPoolBase> GetAll()
        {
            return _poolMappings.Values;
        }

        public void Destory(ThreadWorkerPoolBase pool)
        {
            if (pool == null)
                throw new ArgumentNullException("pool");

            lock (_createLocker)
            {
                if (!_poolMappings.ContainsKey(pool.Name))
                    throw new ArgumentException("Will destory ThreadWorkerPool don't belong current ThreadWorkerPoolManager!", "pool");

                string poolName = pool.Name;
                var destoryPool = (ThreadWorkerPoolBase)pool;
                destoryPool.Destory();
                _poolMappings.Remove(poolName);
            }
        }

        private bool _isDisposed = false;

        public void Dispose()
        {
            if (!_isDisposed)
            {
                List<ThreadWorkerPoolBase> poolList = new List<ThreadWorkerPoolBase>(_poolMappings.Values);
                foreach (ThreadWorkerPoolBase pool in poolList)
                {
                    this.Destory(pool);
                }
            }
        }
    }
}
